/**********************************
 monome-DS          by GrizzlyAdams


 based on:

 KAOS-DS            by 0xtob


Source code licensed under the GPL.
***********************************/

#include <nds.h>
#include <nds/arm9/console.h>

#include <stdio.h>
#include <stdarg.h>

#include "libdsmi.h"

#include "logoimg_bin.h"
#include "spriteimg_bin.h"

#define PEN_DOWN (~IPC->buttons & (1 << 6))

int monome_connect = 0;
int monome_connected = 0;

touchPosition touch;
int touch_was_down = 0;

static SpriteEntry sprites[128];
//static pSpriteRotation spriteRotations = (pSpriteRotation)(void*)sprites;

int sprites_dirty = 1;

u8 channel = 0;

#define TXR(x,y) ((((y) & 0xfc) << 1) | (((x) & 0x78) >> 3))

void sgIP_dbgprint(char * txt, ...) {
        char buffer[256];
        va_list args;
        va_start(args,txt);
        vsprintf(buffer,txt,args);
        iprintf(buffer);
}

void monome_event(u8 x, u8 y, u8 p)
{
	if(monome_connected) {
		dsmi_write(0x80 | channel | (p<<4), y*8+x, 127);
	} else {
		sprites[y*8+x].attribute[2] = ATTR2_ALPHA(1) | ATTR2_PRIORITY(1) | TXR(p?32:0,0);
		sprites_dirty = 1;
	}		
}

// Send all notes off
void monome_x()
{
	if(monome_connected) {
		dsmi_write(0xB0 | channel, 123, 0x00);
	} else {
		int i;
		for(i = 0; i<64; i++)
			sprites[i].attribute[2] = ATTR2_ALPHA(1) | ATTR2_PRIORITY(1) | TXR(0,0);
		sprites_dirty = 1;
	}
}

void monome_y()
{
//	dsmi_write(0xB0 | channel, 0x01, 0x00);
}

u8 lastchannel = 255;
void printChannel()
{
	if(lastchannel < 16)
		sprites[64+lastchannel].attribute[2] = ATTR2_ALPHA(1) | ATTR2_PRIORITY(1) | TXR(0,32);
	if(channel < 16)
		sprites[64+channel].attribute[2] = ATTR2_ALPHA(1) | ATTR2_PRIORITY(1) | TXR(32,32);
	sprites_dirty = 1;
}

// Copy sprites to OAM
void updateOAM(void *Source, void *Dest)
{
        DC_FlushRange(Source,128*sizeof(SpriteEntry));
        dmaCopy(Source, Dest, 128 * sizeof(SpriteEntry));
}


u8 fx = 255;
u8 lastx = 255;
u8 lasty = 255;
void VblankHandler()
{
	int i;

	if(monome_connected) {
		// Read upto 16 events
		for(i = 0; i<16; i++) {
			u8 m,a,b;
			if(dsmi_read(&m, &a, &b) == 0) break;
			if((m & 0xf) == channel) {
				if(a<64)
				switch(m >> 4) {
					case 0x8:
						sprites[a].attribute[2] = ATTR2_ALPHA(1) | ATTR2_PRIORITY(1) | TXR(0,0);
						sprites_dirty = 1;
						break;
					case 0x9:
						sprites[a].attribute[2] = ATTR2_ALPHA(1) | ATTR2_PRIORITY(1) | TXR(32,0);
						sprites_dirty = 1;
						break;
				}
			}
		}
	}

	// Update channel sprites if necessary
	if(channel != lastchannel) {
		printChannel();
		lastchannel = channel;
	}

	// Update OAM if necessary
	if(sprites_dirty) {
		updateOAM(sprites, OAM_SUB);
		sprites_dirty = 0;
	}

	scanKeys();
	touch = touchReadXY();
	u8 x = touch.px / 24;
	u8 y = touch.py / 24;

	if(touch.px > 207) {
		u8 z = touch.px - 208;
		x = (z / 24) + 9;
#if 0
		if(((z % 24) < 3) || ((z % 24) > 21) ||
			((touch.py % 24) < 3) || ((touch.py % 24) > 21)) {
			x = 8;
		}
	} else
	if(((touch.px % 24) < 3) || ((touch.px % 24) > 21) ||
		((touch.py % 24) < 3) || ((touch.py % 24) > 21)) {
		x = 8;
#endif
	}

	
	if(!touch_was_down && PEN_DOWN) {
		fx = x;
		touch_was_down = 1;

		if(x < 8) {
			monome_event(x, y, 1);
		} else if(x > 8) {
			channel = (x-9)*8+y;
		}
	}
	else if(touch_was_down && !PEN_DOWN)
	{
		if((fx <= 8) && (lastx < 8)) {
			monome_event(lastx, lasty, 0);
		}
		touch_was_down = 0;
	}

	if(touch_was_down && PEN_DOWN && ((lastx != x) || (lasty != y))) {
		if(fx <= 8) {
			if(lastx < 8)
				monome_event(lastx, lasty, 0);
			if(x < 8)
				monome_event(x, y, 1);
		} else if((fx > 8) && (x > 8)) {
			channel = (x-9)*8+y;
		}
	}

	lastx = x;
	lasty = y;

	u16 keys = keysDown();
	
	if(keys & KEY_X)
	{
		monome_x();
	}
	
	if(keys & KEY_Y)
	{
		monome_y();
	}

	if(keys & KEY_UP)
	{
		if(channel < 15)
			channel++;
	}
	
	if(keys & KEY_DOWN)
	{
		if(channel > 0)
			channel--;
	}

	if(keys & KEY_START)
	{
		if((monome_connect == 0) && (monome_connected == 0))
			monome_connect = 1;
	}

}

int main(void)
{
	u16 i;
	powerON(POWER_ALL);
	irqInit();
	
	lcdMainOnTop();
	
	irqEnable(IRQ_VBLANK);

	videoSetMode(MODE_5_2D);
	videoSetModeSub(MODE_5_2D);

	for(i=0; i<128; i++) {
		sprites[i].attribute[0] = ATTR0_DISABLED;
		sprites[i].attribute[1] = 0;
		sprites[i].attribute[2] = 0;
		sprites[i].attribute[3] = 0;
	}
	updateOAM(sprites, OAM_SUB);

	// Set banks
	VRAM_A_CR = VRAM_ENABLE | VRAM_A_MAIN_BG_0x06000000;
	VRAM_B_CR = VRAM_ENABLE | VRAM_B_MAIN_BG_0x06020000;
	VRAM_C_CR = VRAM_ENABLE | VRAM_C_MAIN_BG_0x06040000;
	VRAM_D_CR = VRAM_ENABLE | VRAM_D_SUB_SPRITE;
	

	// Gfx on main
	BG3_CR = BG_BMP16_256x256 | BG_MAP_BASE(8) | BG_PRIORITY(1);
	BG3_XDX = 1 << 8;
	BG3_XDY = 0;
	BG3_YDX = 0;
	BG3_YDY = 1 << 8;

	for(i=0; i<256*128; i++)
		BG_GFX[i+(256*256)] = 0x8000;

	u16 *logoimg_u16 = (u16 *)logoimg_bin;
	for(i=0; i<256*64; i++)
		BG_GFX[i+(256*384)] = logoimg_u16[i];

	u16 *spriteimg_u16 = (u16 *)spriteimg_bin;
	for(i=0; i<128*96; i++)
		SPRITE_GFX_SUB[i] = (spriteimg_u16[i] == 0xfc1f) ? 0 : spriteimg_u16[i];

	// Text bg on main
	BG0_CR = BG_MAP_BASE(4) | BG_TILE_BASE(0) | BG_PRIORITY(0);
	BG_PALETTE[255] = RGB15(31,15,0);
	consoleInitDefault((u16*)SCREEN_BASE_BLOCK(4), (u16*)CHAR_BASE_BLOCK(0), 16);

	// Monome Buttons
	for(i=0; i<64; i++) {
		sprites[i].attribute[0] = ATTR0_BMP | ((i/8)*24);
		sprites[i].attribute[1] = ATTR1_SIZE_32 | ((i%8)*24);
		sprites[i].attribute[2] = ATTR2_ALPHA(1) | ATTR2_PRIORITY(1) | TXR(0,0);
	}

	// Channel Buttons
	for(i=0; i<16; i++) {
		sprites[64+i].attribute[0] = ATTR0_BMP | ((i%8)*24);
		sprites[64+i].attribute[1] = ATTR1_SIZE_32 | (208+((i/8)*24));
		sprites[64+i].attribute[2] = ATTR2_ALPHA(1) | ATTR2_PRIORITY(1) | TXR(0,32);
	}

	// Separator Bar
	for(i=0; i<2; i++) {
		// Middle
		sprites[80+i].attribute[0] = ATTR0_BMP | ATTR0_TALL | (i*64+32);
		sprites[80+i].attribute[1] = ATTR1_SIZE_64 | (192);
		sprites[80+i].attribute[2] = ATTR2_ALPHA(1) | ATTR2_PRIORITY(1) | TXR(96,0);
	}

	// Top
	sprites[82].attribute[0] = ATTR0_BMP | ATTR0_TALL | (0);
	sprites[82].attribute[1] = ATTR1_SIZE_32 | (192);
	sprites[82].attribute[2] = ATTR2_ALPHA(1) | ATTR2_PRIORITY(1) | TXR(64,0);

	// Bottom
	sprites[83].attribute[0] = ATTR0_BMP | (160);
	sprites[83].attribute[1] = ATTR1_SIZE_16 | (192);
	sprites[83].attribute[2] = ATTR2_ALPHA(1) | ATTR2_PRIORITY(1) | TXR(64,32);

	// Status
	sprites[84].attribute[0] = ATTR0_BMP | (176);
	sprites[84].attribute[1] = ATTR1_SIZE_16 | (192);
	sprites[84].attribute[2] = ATTR2_ALPHA(1) | ATTR2_PRIORITY(1) | TXR(80,32);

	// Channel Numbers
	for(i=0; i<16; i++) {
		sprites[85+i].attribute[0] = ATTR0_BMP | ((i%8)*24+4);
		sprites[85+i].attribute[1] = ATTR1_SIZE_16 | ((i/8)*24+212);
		sprites[85+i].attribute[2] = ATTR2_ALPHA(1) | ATTR2_PRIORITY(0) | TXR((i%8)*16,(i/8)*16+64);
	}

	// Set modes
	videoSetMode(MODE_5_2D | DISPLAY_BG0_ACTIVE | DISPLAY_BG3_ACTIVE);
	videoSetModeSub(MODE_5_2D | DISPLAY_SPR_ACTIVE | DISPLAY_SPR_2D | DISPLAY_SPR_2D_BMP_128);

	dsmi_setup_wifi_support();

	monome_connected = 0;
	monome_connect = 0;

	iprintf("\x1b[0;0H\x1b[K monome-DS v1.0 by GrizzlyAdams \n");
	iprintf("\x1b[8;0H\x1b[K     Press START to connect.    \n");

	while(1)
	{
		if(monome_connect && !monome_connected) {
			sprites[84].attribute[2] = ATTR2_ALPHA(1) | ATTR2_PRIORITY(1) | TXR(80,0);
			swiWaitForVBlank();
			updateOAM(sprites, OAM_SUB);
			iprintf("\x1b[8;0H\x1b[K          Connecting...\n");

			int res = dsmi_connect();
			if(res == 1) {
				iprintf("\x1b[8;0H\x1b[K           Connected.\n");
				sprites[84].attribute[2] = ATTR2_ALPHA(1) | ATTR2_PRIORITY(1) | TXR(80,16);
				monome_connected = 1;
			} else {
				iprintf("\x1b[8;0H\x1b[K    Oh no, could not connect!\n");
				sprites[84].attribute[2] = ATTR2_ALPHA(1) | ATTR2_PRIORITY(1) | TXR(80,32);
				monome_connected = 0;
			}
			sprites_dirty = 1;
#if 0
		} else if(!monome_connect && monome_connected) {
			dsmi_disconnect();
			monome_connected = 0;
#endif
		}

		swiWaitForVBlank();
		VblankHandler();
	}
	
}
